package mhtree.benchmarking;

import messif.buckets.BucketStorageException;
import messif.buckets.impl.MemoryStorageBucket;
import messif.objects.LocalAbstractObject;
import messif.operations.data.InsertOperation;
import messif.operations.query.ApproxKNNQueryOperation;
import mhtree.InsertType;
import mhtree.MHTree;

import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

public class RunBenchmark {
    private static List<String> datasets = Arrays.asList(
            "./data/mix_m.txt",
            "./data/intersection.txt",
            "./data/middle_c.txt",
            "./data/r1.txt"
    );

    public static void main(String[] args) {
        if (args.length != 0) datasets = Arrays.asList(args.clone());

        new BenchmarkConfig(
                Arrays.asList(
                        new MHTreeConfig(500, 5, 16, InsertType.GREEDY),
                        new MHTreeConfig(500, 5, 16, InsertType.INCREMENTAL)
                ),
                datasets,
                Arrays.asList(
                        RunBenchmark::benchBuildTree,
                        (MHTreeConfig config) -> benchApproxKNN(config, 30),
                        RunBenchmark::benchInsert
                )).execute();
    }

    public static void benchInsert(MHTreeConfig config) {
        try {
            List<LocalAbstractObject> objects = config.getObjects();
            List<LocalAbstractObject> initialObjects = objects.subList(0, objects.size() / 2);
            List<LocalAbstractObject> restOfTheObjects = objects.subList(objects.size() / 2, objects.size());

            long timeStart = System.currentTimeMillis();

            MHTree tree = new MHTree(
                    initialObjects,
                    config.getLeafCapacity(),
                    config.getNumberOfChildren(),
                    config.getNumberOfThreads(),
                    config.getInsertType(),
                    MemoryStorageBucket.class,
                    null);

            long treeBuilt = System.currentTimeMillis();

            long tookMax = Long.MIN_VALUE;
            long tookMin = Long.MAX_VALUE;

            for (LocalAbstractObject object : restOfTheObjects) {
                long insertStart = System.currentTimeMillis();
                tree.insert(new InsertOperation(object));
                long took = System.currentTimeMillis() - insertStart;

                tookMax = Math.max(tookMax, took);
                tookMin = Math.min(tookMin, took);
            }

            long timeStop = System.currentTimeMillis();

            System.out.println("~Building took: " + (treeBuilt - timeStart) + " ms");
            System.out.println("~max: " + tookMax + " ms");
            System.out.println("~min: " + tookMin + " ms");
            System.out.println("~Took: " + (timeStop - timeStart) + " ms");

            tree.printStatistics();
        } catch (BucketStorageException ex) {
            Logger.getLogger("BuildTree").log(Level.SEVERE, null, ex);
        }
    }

    public static void benchApproxKNN(MHTreeConfig config, int k) {
        try {
            MHTree tree = new MHTree(config.getObjects(),
                    config.getLeafCapacity(),
                    config.getNumberOfChildren(),
                    config.getNumberOfThreads(),
                    config.getInsertType(),
                    MemoryStorageBucket.class,
                    null);

            long timeStart = System.currentTimeMillis();

            long tookMax = Long.MIN_VALUE;
            long tookMin = Long.MAX_VALUE;

            for (LocalAbstractObject object : config.getObjects()) {
                long nnStart = System.currentTimeMillis();

                tree.approxKNN(new ApproxKNNQueryOperation(object, k));

                long took = System.currentTimeMillis() - nnStart;

                tookMax = Math.max(tookMax, took);
                tookMin = Math.min(tookMin, took);
            }

            long timeStop = System.currentTimeMillis();

            System.out.println("~nn took: " + (timeStop - timeStart) + " ms");
            System.out.println("~max: " + tookMax + " ms");
            System.out.println("~min: " + tookMin + " ms");

            System.out.println("~" + config + " k: " + k);

            tree.printStatistics();
        } catch (BucketStorageException ex) {
            Logger.getLogger("BuildTree").log(Level.SEVERE, null, ex);
        }
    }

    public static void benchBuildTree(MHTreeConfig config) {
        try {
            long timeStart = System.currentTimeMillis();
            MHTree tree = new MHTree(
                    config.getObjects(),
                    config.getLeafCapacity(),
                    config.getNumberOfChildren(),
                    config.getNumberOfThreads(),
                    config.getInsertType(),
                    MemoryStorageBucket.class,
                    null);
            long timeStop = System.currentTimeMillis();

            System.out.println("~Building took: " + (timeStop - timeStart) + " ms");

            tree.printStatistics();
        } catch (BucketStorageException ex) {
            Logger.getLogger("BuildTree").log(Level.SEVERE, null, ex);
        }
    }
}
